# copyright (c) 2018- polygoniq xyz s.r.o.

import typing
import bpy
import os
import glob
import hashlib
import functools
import dataclasses
import logging
from . import derivative_generator
logger = logging.getLogger(f"polygoniq.{__name__}")


ORIGINAL_PATH_PROPERTY_NAME = "memsaver_original_path"
# TODO: change value appropriately, test whether the change causes an issue
DERIVATIVE_FILENAME_PROPERTY_NAME = "memsaver_derivative_path"
DERIVATIVE_SIZE_PROPERTY_NAME = "memsaver_derivative_size"


def remove_memsaver_properties(image: bpy.types.Image) -> None:
    image[DERIVATIVE_FILENAME_PROPERTY_NAME] = None
    image[ORIGINAL_PATH_PROPERTY_NAME] = None
    image[DERIVATIVE_SIZE_PROPERTY_NAME] = 0


def was_filepath_overwritten_on_derivative(image: bpy.types.Image) -> bool:
    """Verifies whether an image derivative is pointing to a different file then memsaver expects.

    If the image derivative's filepath is pointing to a different image then expected,
    it was changed outside of memsaver. Such image can no longer be considered an image derivative.
    """
    # image is generated or packed, memsaver does not handle the filepath
    if image.filepath in {None, ""}:
        return False

    # the image is not an image derivative, memsaver does not handle the filepath
    if image.get(DERIVATIVE_FILENAME_PROPERTY_NAME, None) is None:
        return False

    return image[DERIVATIVE_FILENAME_PROPERTY_NAME].lower() != \
        bpy.path.basename(image.filepath).lower()


def get_filepath_hash(abs_path: str) -> str:
    """Given an absolute file path, return a hash

    This hash is the base for image derivative paths.
    """
    # TODO: We generate the hash from absolute path, is that OK?
    assert os.path.isabs(abs_path)
    return hashlib.sha256(abs_path.encode("utf-8")).hexdigest()


def get_original_path(image: bpy.types.Image) -> typing.Optional[str]:
    # original path set by memsaver is incorrect if the derivative's filepath was changed
    # outside of memsaver, so there is no original path to get
    assert not was_filepath_overwritten_on_derivative(image)

    original_path = image.get(ORIGINAL_PATH_PROPERTY_NAME, None)
    if original_path is not None:
        assert isinstance(original_path, str)
        return typing.cast(str, original_path)
    return None


def ensure_original_path(image: bpy.types.Image) -> str:
    if was_filepath_overwritten_on_derivative(image):
        remove_memsaver_properties(image)
    original_path = get_original_path(image)
    if original_path in {None, ""}:
        # we don't absolutize the path, we will do that only just before we have to
        image[ORIGINAL_PATH_PROPERTY_NAME] = image.filepath
    else:
        return typing.cast(str, original_path)

    original_path = get_original_path(image)
    # it can be "" in case the image is generated or packed
    assert original_path is not None
    return typing.cast(str, original_path)


def revert_to_original(image: bpy.types.Image) -> None:
    if was_filepath_overwritten_on_derivative(image):
        remove_memsaver_properties(image)
        logger.debug(
            f"Asked to revert image {image.name} to the original but it's filepath was changed "
            f"outside of memsaver, filepath: {image.filepath}.")
        return

    original_path = get_original_path(image)
    if original_path in {None, ""}:
        remove_memsaver_properties(image)
        logger.debug(
            f"Asked to revert image {image.name} to the original but it already is, "
            f"filepath: {image.filepath}.")
        return

    image.filepath = original_path
    logger.debug(f"Reverted image {image.name} to the original at {original_path} (bpy path).")


def is_derivative_stale(original_abs_path: str, derivative_path: str) -> bool:
    """Returns True if derivative image does not exist or the original image has changed
    """
    if not os.path.isfile(derivative_path):
        logger.debug(
            f"Derivative at path {derivative_path} is stale because it doesn't exist!")
        return True  # doesn't exist, must be stale

    # derivative has been modified earlier than original, must be stale
    if os.path.getmtime(derivative_path) <= os.path.getmtime(original_abs_path):
        logger.debug(
            f"Derivative at path {derivative_path} is stale because the original "
            f"{original_abs_path} has a newer modified date!")
        return True

    logger.debug(
        f"Derivative at path {derivative_path} is up to date, original at {original_abs_path}.")
    return False


def find_sequence_num_indices(filename: str) -> typing.Tuple[int, int]:
    """Returns tuple (num_start, num_end) with indices of the first (last) digit of sequence number.

    Returns (-1, -1) if the filename doesn't contain sequence number.
    """
    # Blender is very robust when considering sequence numbers, it iterates through filename
    # from end to start and the first numeric sequence found is considered to be the sequence number:
    # https://github.com/blender/blender/blob/36983fb5e4a297cb26855b6777a7ce25c12d7c49/source/blender/blenlib/intern/path_util.c#L70
    num_start, num_end = -1, -1
    found_digit = False
    for i in range(len(filename) - 1, -1, -1):
        if filename[i].isdigit():
            if found_digit:
                num_start = i
            else:
                found_digit = True
                num_start, num_end = i, i
        elif found_digit:
            break

    return (num_start, num_end)


def udim_matcher(filename: str, start_index: int) -> typing.Optional[typing.Tuple[str, str]]:
    """Tries to match UDIM number in 'filename' starting at 'start_index'.

    UDIM is one format for UDIM/Tiles textures, it has format: 1001 + u-tile + v-tile * 10   (e.g. 1012)
    Returns matched UDIM number and filename without UDIM or None if UDIM wasn't found.
    """
    candidate = filename[start_index: start_index + 4]
    if len(candidate) < 4:
        return None
    if not candidate.isdigit():
        return None
    return (candidate, filename[:start_index] + filename[start_index + 4:])


def uvtile_matcher(filename: str, start_index: int) -> typing.Optional[typing.Tuple[str, str]]:
    """Tries to match UV Tile string in 'filename' starting at 'start_index'.

    UV Tile is one format for UDIM/Tiles textures, it has format: u(u-tile + 1)_v(v-tile + 1)   (e.g. u1_v2)
    Returns matched UV Tile string and filename without UV Tile or None if UV Tile wasn't found.
    """
    candidate = filename[start_index: start_index + 5]
    if len(candidate) < 5:
        return None
    if candidate[0] == "u" and candidate[1].isdigit() and candidate[2] == "_"\
            and candidate[3] == "v" and candidate[4].isdigit():
        return (candidate, filename[:start_index] + filename[start_index + 5:])
    return None


@dataclasses.dataclass
class PathMapping:
    """Mapping from an original to a derivate path.

    Simple container with named items, used dataclass instead of NamedTuple to be consistent with
    OrigToDerivativePaths.
    """
    original: str
    derivative: str


@dataclasses.dataclass
class OrigToDerivativePaths:
    """Dataclass for storing mapping from original to derivative paths for one image datablock.

    'assigned_path' is mapping from path that is originally assigned in image.filepath to the
    new path that should be assigned to image.filepath

    'resource_paths' are mappings from actual image files on disc to the derivative files on the disc

    These two are different for UDIMs where image.filepath has assigned path with <UDIM> or <UVTILE>
    token which doesn't correspond to any file on disc. Blender then references all image files
    matched by those tokens.
    """
    assigned_path: PathMapping
    resource_paths: typing.List[PathMapping]


def get_derivative_paths(
    image: bpy.types.Image,
    cache_path: str,
    original_abs_path: str,
    side_size: int
) -> typing.Optional[OrigToDerivativePaths]:
    """Returns mapping from the original to the derivative image paths.

    None is returned if we don't support resizing image of given type or derivative paths cannot be
    safely inferred.
    """
    dir_path = os.path.dirname(original_abs_path)
    filename, ext = os.path.splitext(os.path.basename(original_abs_path))

    if image.source == 'FILE':
        # Simple single image on the disc
        path_hash = get_filepath_hash(original_abs_path)
        derivative_path = os.path.join(cache_path, f"{path_hash}_{side_size}{ext}")
        path_map = PathMapping(original_abs_path, derivative_path)
        return OrigToDerivativePaths(path_map, [path_map])
    elif image.source == 'SEQUENCE':
        # Sequence of images, one of the images is assigned to image.filepath, Blender then finds
        # the rest of the images based on varying sequence number on the name.
        # e.g. image_0001_diffuse.png -> image_0002_diffuse.png, image_0003_diffuse.png, ..
        num_start, num_end = find_sequence_num_indices(filename)
        if num_start < 0 or num_end < 0:
            logger.error(
                f"Image {image.name} with filepath {image.filepath} of type 'SEQUENCE' does not "
                f"have a sequence number in the filename, skipping derivates generation!"
            )
            return None

        # Generate derivative path for the assigned image
        main_digit = filename[num_start:num_end + 1]
        orig_filename_wo_digit = filename[:num_start] + filename[num_end + 1:]
        path_hash = get_filepath_hash(os.path.join(dir_path, orig_filename_wo_digit + ext))
        derivative_path = os.path.join(cache_path, f"{path_hash}_{side_size}_{main_digit}{ext}")
        path_map = PathMapping(original_abs_path, derivative_path)
        result = OrigToDerivativePaths(path_map, [path_map])

        # Generate derivative paths for all the corresponding sequence images found on discs
        for potential_seq_filepath in glob.glob(f"{dir_path}/*{ext}"):
            potential_seq_filename, _ = os.path.splitext(os.path.basename(potential_seq_filepath))
            digit = potential_seq_filename[num_start:num_end + 1]
            filename_wo_digit = potential_seq_filename[:num_start] + \
                potential_seq_filename[num_end + 1:]
            if not digit.isdigit() or digit == main_digit or filename_wo_digit != orig_filename_wo_digit:
                continue
            derivative_path = os.path.join(cache_path, f"{path_hash}_{side_size}_{digit}{ext}")
            result.resource_paths.append(PathMapping(potential_seq_filepath, derivative_path))
        return result
    elif image.source == 'TILED':
        # Find type of token and it's starting index
        if (udim_start := filename.find("<UDIM>")) > -1:
            token = "<UDIM>"
            orig_filename_wo_token = filename.replace(token, "", 1)
            matcher = functools.partial(udim_matcher, start_index=udim_start)
        elif (udim_start := filename.find("<UVTILE>")) > -1:
            token = "<UVTILE>"
            orig_filename_wo_token = filename.replace(token, "", 1)
            matcher = functools.partial(uvtile_matcher, start_index=udim_start)

        if udim_start == -1:
            logger.error(
                f"Image {image.name} with filepath {image.filepath} of type 'TILED' does not have "
                f"<UDIM> nor <UVTILE> tag in the filename, skipping derivates generation!"
            )
            return None

        # Generate derivative path for filepath with token
        path_hash = get_filepath_hash(os.path.join(dir_path, orig_filename_wo_token + ext))
        derivative_path = os.path.join(cache_path, f"{path_hash}_{side_size}.{token}{ext}")
        path_map = PathMapping(original_abs_path, derivative_path)
        result = OrigToDerivativePaths(path_map, [])

        # Search for TILED images on disc and generate derivative path for them
        for potential_tile_filepath in glob.glob(os.path.join(dir_path, f"*{ext}")):
            potential_tile_filename, _ = os.path.splitext(os.path.basename(potential_tile_filepath))
            split = matcher(potential_tile_filename)
            if split is None:
                continue
            tile_num, filename_wo_token = split
            if orig_filename_wo_token != filename_wo_token:
                continue
            derivative_path = os.path.join(cache_path, f"{path_hash}_{side_size}.{tile_num}{ext}")
            result.resource_paths.append(PathMapping(potential_tile_filepath, derivative_path))
        return result
    else:  # One of 'MOVIE', 'GENERATED', 'VIEWER'
        logger.debug(
            f"Image '{image.name}' is of type '{image.source}' and thus no derivative will be "
            f"generated. We don't support it yet or it doesn't even make sense to resize!"
        )
        return None


def ensure_and_assign_derivative_images(
        cache_path: str, original_path: str, image: bpy.types.Image, side_size: int) -> None:
    """Generate derivatives of the image and assign them to the image object
    """
    original_abs_path = os.path.abspath(bpy.path.abspath(original_path, library=image.library))
    image_paths = get_derivative_paths(image, cache_path, original_abs_path, side_size)
    if image_paths is None:
        return

    error_occured = False
    for path_map in image_paths.resource_paths:
        orig_path, derivative_path = path_map.original, path_map.derivative
        if not is_derivative_stale(orig_path, derivative_path):
            logger.debug(
                f"Derivative {derivative_path} of size {side_size} from original {orig_path} is "
                f"up to date, no need to re-generate."
            )
            continue

        if os.path.isfile(derivative_path):
            os.unlink(derivative_path)
        error_occured = not derivative_generator.generate_derivative(
            orig_path,
            derivative_path,
            side_size
        )
        if error_occured:
            logger.error(
                f"Failed to generate derivative {derivative_path} of size {side_size} "
                f"from original {orig_path}."
            )
            # Break here, we'll keep the current image
            break
        else:
            logger.debug(
                f"Generated derivative {derivative_path} of size {side_size} from original "
                f"{orig_path} because it was stale."
            )

    if not error_occured:
        image.filepath = image_paths.assigned_path.derivative
        # duplicating information to be able to tell if image was changed outside of memsaver
        # saving a basename of the derivative should be enough because it contains the hash suffix
        image[DERIVATIVE_FILENAME_PROPERTY_NAME] = bpy.path.basename(image.filepath)
        image[DERIVATIVE_SIZE_PROPERTY_NAME] = side_size
        logger.debug(
            f"Set image {image.name} to derivative {image_paths.assigned_path.derivative} of "
            f"size {side_size}, original at {image_paths.assigned_path.original}."
        )
    else:
        if all(os.path.isfile(path_map.original) for path_map in image_paths.resource_paths):
            # We failed to generate the derivative, which means we could be stuck on outdated or
            # some weird derivative from the past, e.g. 1x1.
            # Switch to original as all original images exist.
            revert_to_original(image)
            logger.warning(
                f"Failed to generate derivatives from original {original_abs_path}, "
                f"reverted image to the original!"
            )
        else:
            logger.warning(
                f"Failed to generate derivatives from original {original_abs_path} "
                f"and original doesn't exist, keeping old derivative!"
            )


def change_image_size(cache_path: str, image: bpy.types.Image, side_size: int) -> None:
    assert side_size > 0
    original_path = ensure_original_path(image)
    if original_path == "":
        logger.warning(
            f"Image {image.name} has empty original path, it's most likely generated or packed. "
            f"Can't change its size."
        )
        return

    if image.packed_file is not None:
        logger.warning(f"Image {image.name} is packed. Can't change its size.")
        return

    if image.get(DERIVATIVE_SIZE_PROPERTY_NAME, 0) == 0:
        # if the image has the original path and its size is smaller than what's requested, we can
        # do an early out optimization here
        if max(image.size) <= side_size:
            logger.debug(
                f"Image {image.name} has original size {list(image.size)} which is smaller or "
                f"equal to the requested side size {side_size}. Skipping!"
            )
            return

    ensure_and_assign_derivative_images(cache_path, original_path, image, side_size)


def check_derivative(cache_path: typing.Optional[str], image: bpy.types.Image) -> None:
    """Re-generate derivative image if the original image file was changed on disk

    cache_path may be None, in this case it will be inferred from derivative absolute path
    """
    if was_filepath_overwritten_on_derivative(image):
        remove_memsaver_properties(image)
        return

    original_path = get_original_path(image)
    if original_path is None:  # it's original or original path custom property is not present
        return

    side_size: int = image.get(DERIVATIVE_SIZE_PROPERTY_NAME, 0)
    if side_size == 0:  # it's original or no custom property is present
        return

    # we can infer cache_path, this is useful especially in the post_load handler
    if cache_path is None:
        cache_path = os.path.dirname(image.filepath)
        if not os.path.isdir(cache_path):
            logger.warning(
                f"Inferred cache_path {cache_path} from image.filepath {image.filepath}, "
                f"however the cache_path is not a valid directory! Skipping!")
            return

    ensure_and_assign_derivative_images(cache_path, original_path, image, side_size)
